mod helper;
pub mod layout;
pub mod file_map;

use cdtypes::types::{cdstring::DString, dir_record::DirectoryRecord, path_table::PathTableL};
use file_map::FileSystemMap;
use std::{cell::RefCell, rc::Rc};

pub use tool_helper::Error;
pub type SharedPtr<T> = Rc<RefCell<T>>;

pub fn new_shared_ptr<T>(value: T) -> SharedPtr<T> {
    Rc::new(RefCell::new(value))
}

pub struct CDDesc {
    pub(super) system_area:      SharedPtr<SystemArea>,
    pub(super) path_table:       SharedPtr<PathTable>,
    pub(super) pvd:              SharedPtr<PrimaryVolumeDescriptor>,
    pub(super) root:             SharedPtr<Directory>,
    pub(super) vol_sector_count: usize,
}

impl CDDesc {
    pub fn new() -> CDDesc {
        match Directory::new("\x00") {
            Ok(root)   => CDDesc{system_area: new_shared_ptr(SystemArea::new()), path_table: new_shared_ptr(PathTable::new()), pvd: new_shared_ptr(PrimaryVolumeDescriptor::new()), root: new_shared_ptr(root), vol_sector_count: 0},
            Err(error) => panic!("Creating root directory failed with: {}", error)
        }
    }

    pub fn get_memory_layout(&self) -> layout::MemoryLayout {
        layout::DefaultLayout::new(self)
    }

    pub fn add_dir(&mut self, dir: Directory) {
        self.root.borrow_mut().add_dir(dir);
    }

    pub fn add_file(&mut self, file: File) {
        self.root.borrow_mut().add_file(file);
    }

    pub fn create_file_map(&self) -> FileSystemMap {
        file_map::new_file_map(&self.root.borrow())
    }

    pub(super) fn for_each_dir_mut<T: Fn(&mut Directory)>(dir: SharedPtr<Directory>, function: &T) {
        let mut dir = dir.borrow_mut();

        function(&mut dir);
        for sub_dir in &mut dir.dirs {
            Self::for_each_dir_mut(sub_dir.clone(), function);
        }
    }
}

pub struct SystemArea {
    pub(in super) track_rel_lba: usize,
}

impl SystemArea {
    pub fn new() -> SystemArea {
        SystemArea{track_rel_lba: 0}
    }
}

pub struct PathTable {
    pub(super) track_rel_lba: usize,
    pub(super) size_bytes:    usize,
}

impl PathTable {
    pub fn new() -> PathTable {
        PathTable{track_rel_lba: 0, size_bytes: 0}
    }

    pub fn calculate_size_for(root: SharedPtr<Directory>) -> usize {
        let mut size_bytes = 0;

        helper::collect_path_table_member(root).into_iter().for_each(|element| {
            println!("PT: {} ^{}", element.name, element.parent_table_id);
            size_bytes += PathTableL::calculate_size_for(element.name.as_ref());
        });

        size_bytes
    }

    pub fn get_track_rel_lba_for(&self, table_num: usize, sector_count_func: fn(data_size: usize) -> usize) -> usize {
        let table_num = table_num - 1;
        self.track_rel_lba + (table_num*sector_count_func(self.size_bytes))
    }
}

pub struct PrimaryVolumeDescriptor {
    pub(super) track_rel_lba: usize,
    pub(super) publisher:     String,
}

impl PrimaryVolumeDescriptor {
    pub fn new() -> PrimaryVolumeDescriptor {
        PrimaryVolumeDescriptor{track_rel_lba: 0, publisher: String::new()}
    }
}

pub struct Directory {
    pub name:       DirectoryName,
    pub properties: Properties,
        size_bytes: usize,
        files:      Vec<SharedPtr<File>>,
        dirs:       Vec<SharedPtr<Directory>>
}

impl Directory {
    pub fn new(dir_name: &str) -> Result<Directory, Error> {
        Ok(Directory{name: DirectoryName::from_str(dir_name)?, properties: Properties::default(), size_bytes: 0, files: Vec::new(), dirs: Vec::new()})
    }

    pub fn add_dir(&mut self, dir: Directory) {
        self.dirs.push(new_shared_ptr(dir));
    }

    pub fn add_file(&mut self, file: File) {
        self.files.push(new_shared_ptr(file));
    }

    pub fn get_track_rel_lba(&self) -> usize {
        self.properties.track_rel_lba
    }

    pub fn get_size(&self) -> usize {
        self.properties.get_size(self.size_bytes)
    }

    pub(super) fn update_content_size(&mut self) {
        let dir_member     = helper::collect_directory_record_member(&self.files, &self.dirs);
        let mut size_bytes = 0;

        for member in dir_member {
            size_bytes += DirectoryRecord::calculate_size_for(member.get_name().as_str(), true);
        }

        self.size_bytes = size_bytes;
    }
}

impl std::fmt::Display for Directory {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Directory: {}", self.name)
    }    
}

enum FileType {
    Regular(Vec<u8>)
}

pub struct File {
    pub name:       FileName,
    pub properties: Properties,
        content:    FileType
}

impl File {
    pub fn new_regular(file_name: &str, content: Vec<u8>) -> Result<File, Error> {
        Ok(File{name: FileName::from_str(file_name)?, properties: Properties::default(), content: FileType::Regular(content)})
    }

    pub fn get_track_rel_lba(&self) -> usize {
        self.properties.track_rel_lba
    }

    pub fn get_size(&self) -> usize {
        self.properties.get_size(self.get_content_size())
    }

    pub(super) fn get_content_size(&self) -> usize {
        match &self.content {
            FileType::Regular(content) => content.len()
        }
    }
}

impl std::fmt::Display for File {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "File: {}", self.name)
    }
}

pub struct DirectoryName {
    name: DString<8>,
    len:  usize,
}

impl DirectoryName {
    pub fn from_str(dir_name: &str) -> Result<DirectoryName, Error> {
        let dir_name = dir_name.to_uppercase();
        Ok(DirectoryName{name: DString::from_str(dir_name.as_ref())?, len: dir_name.len()})
    }

    pub fn as_string(&self) -> Option<String> {
        Some(self.as_str()?.to_owned())
    }

    pub fn as_str(&self) -> Option<&str> {
        dstring_as_str(&self.name, self.len)
    }
}

impl std::fmt::Display for DirectoryName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_str().unwrap_or("<???>"))
    } 
}

pub struct FileName {
    name:    DString<8>,
    len:     usize,
    ext:     Option<DString<3>>,
    ext_len: Option<usize>
}

impl FileName {
    pub fn from_str(file_name: &str) -> Result<FileName, Error> {
        let file_name = file_name.to_uppercase();
        let (name, ext) = {
            let mut sub_str = file_name.split('.');
            (Error::ok_or_new(sub_str.next(), || "File name can't be empty".to_owned())?, sub_str.next())
        };
        let (ext, ext_len) = {
            match ext {
                Some(ext) => (Some(DString::from_str(ext)?), Some(ext.len())),
                None      => (None, None)
            }   
        };

        Ok(FileName{name: DString::from_str(name)?, len: name.len(), ext, ext_len})
    }

    pub fn as_string(&self) -> Option<String> {
        if let Some(ext) = &self.ext {
            Some(format!("{}.{};1", dstring_as_str(&self.name, self.len)?, dstring_as_str(&ext, self.ext_len.unwrap_or(0))?))
        }

        else {
            Some(format!("{}", dstring_as_str(&self.name, self.len)?))
        }
    }
}

impl std::fmt::Display for FileName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_string().unwrap_or("<???>".to_owned()))
    }
}

pub struct Properties {
    pub(in super) track_rel_lba:        usize,
    pub(in super) overwrite_size_bytes: Option<usize>,
    pub(in super) is_hidden:            bool
}

impl Properties {
    pub(super) fn get_size(&self, content_size: usize) -> usize {
        if let Some(forced_bytes) = self.overwrite_size_bytes {
            forced_bytes
        }

        else {
            content_size
        }    
    }
}

impl Default for Properties {
    fn default() -> Self {
        Properties{track_rel_lba: 0, overwrite_size_bytes: None, is_hidden: false}
    }
}

fn dstring_as_str<const SIZE: usize>(string: &DString<SIZE>, len: usize) -> Option<&str> {
    match std::str::from_utf8(&string.as_raw()[0..len]) {
        Ok(str) => Some(str),
        Err(_)  => None,
    }
}